/*
 The SWI-Prolog interface to MiniSat SAT solver
 http://www.cs.chalmers.se/Cs/Research/FormalMethods/MiniSat/MiniSat.html

 Copyright (c) 2006, Michael Codish, Vitaly Lagoon, and Peter J. Stuckey
 
 Permission is hereby granted, free of charge, to any person obtaining a
 copy of this software and associated documentation files (the
 "Software"), to deal in the Software without restriction, including
 without limitation the rights to use, copy, modify, merge, publish,
 distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to
 the following conditions:
 
 The above copyright notice and this permission notice shall be included
 in all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/


#include "pl-minisat.h"

static lit clause_buffer[MAX_CLAUSE_LENGTH];
static solver *s=NULL;

static inline lit pl2lit(term_t pl_literal)
{
    int pl_lit_int;
    PL_get_integer(pl_literal,&pl_lit_int);
    return (pl_lit_int<0) ? neg(toLit(-pl_lit_int)) : toLit(pl_lit_int);
}

extern "C" {

foreign_t minisat_new_solver()
{
    s = solver_new();
    PL_succeed;
}

foreign_t minisat_delete_solver()
{
    if (s) {
	solver_delete(s);
	s = NULL;
    }
    PL_succeed;
}

foreign_t minisat_add_clause(term_t l)
{
    //int pl_literal;
    term_t head = PL_new_term_ref();      /* variable for the elements */
    term_t list = PL_copy_term_ref(l);    /* copy as we need to write */
    int *ptr = clause_buffer;
    //bool status;
    while( PL_get_list(list, head, list) ) {
	*ptr = pl2lit(head);
	ptr++;
    }
    
    assert(PL_get_nil(list));
    
    if (solver_addclause(s,clause_buffer,ptr)) PL_succeed;
    else PL_fail;
}

foreign_t minisat_solve(term_t res)
{
    if (!solver_simplify(s)) PL_fail;
    if (!solver_solve(s,NULL,NULL)) PL_fail;

    {
	term_t l = PL_new_term_ref();
	term_t pl_lit = PL_new_term_ref();
	int pl_lit_int;
	int i = solver_nvars(s);
	void **vals = vec_begin(&(s->model));
	
	PL_put_nil(l);
	while( --i >= 0 ) {
	    switch((int)((lbool)vals[i])) {
	    case  1: pl_lit_int = i;  break;
	    case -1: pl_lit_int = -i; break;
	    case  0: pl_lit = 0;
	    default : assert(0);
	    }
	    if (pl_lit_int) {
		PL_put_integer(pl_lit,pl_lit_int);
		PL_cons_list(l, pl_lit, l);
	    }
	}
	return PL_unify(res,l);
    }
}

//=============================================================================
static const PL_extension predicates[] =
    {
        //
        //  { "name", arity, function, PL_FA_<flags> },
        //

        { "minisat_new_solver",     0, minisat_new_solver,      0 },
        { "minisat_delete_solver",  0, minisat_delete_solver,   0 },
	{ "minisat_add_clause",     1, minisat_add_clause,      0 },
	{ "minisat_solve",          1, minisat_solve,           0 },

        { NULL,               0, NULL,             0 }    // terminating line
    };

//-----------------------------------------------------------------------------
install_t install()
{
    //Sdprintf("%% SWI-Prolog interface to MiniSat");
    //Sdprintf(" - built on ");
    //Sdprintf(__DATE__);
    //Sdprintf(", ");
    //Sdprintf(__TIME__);
    //Sdprintf(" ... ");
    PL_register_extensions(predicates);	/* This is the only PL_ call allowed */
					/* before PL_initialise().  It */
					/* ensures the foreign predicates */
					/* are available before loading */
					/* Prolog code */
    
    //Sdprintf("OK\n");
}

} // extern "C"

//-----------------------------------------------------------------------------
// This part is for compiling into a standalone executable

#ifdef READLINE
static void install_readline(int argc, char**argv)
{
    PL_install_readline();
}
#endif

int main(int argc, char **argv)
{

#ifdef READLINE
    PL_initialise_hook(install_readline);
#endif

    install();
    if ( !PL_initialise(argc, argv) )
	PL_halt(1);
    
    PL_halt(PL_toplevel() ? 0 : 1);
    
    return 0;
}
