/*
 * The core group of static methods that analyze a SPoX policy for
 * non-determinism.
 */

package checkpolicy;

import checkpolicy.prolog.*;
import checkpolicy.util.*;
import checkpolicy.sat.*;
import java.util.*;
import java.io.*;

import jpl.Query;
import spox.policy.*;
import spox.policy.pointcut.*;
import spox.policy.PolicyComponent.*;
import spox.policy.pointcut.PointcutComponent.*;

public class Analysis
{

    // Attributes
    private static HashSet<String> usedForallVars = new HashSet<String>();
    private static HashMap<String,String> forallFrom = new HashMap<String,String>();
    private static HashMap<String,String> forallTo = new HashMap<String,String>();

    /**
     * Checks a policy using various methods, including a non-determinism
     * analysis with Prolog.
     * @param policy The policy to analyze.
     * @return "Okay" if the policy checks out; otherwise an explanation of
     * one or more possible issues.
     */
    public static String checkPolicy(Policy policy)
    {


        StringBuilder returnString = new StringBuilder();
        returnString.append(checkForNonDeterminism(policy) + "\n");
        
        return returnString.toString();
    }
    
    /**
     * Analyzes a policy for non-determinism, using Prolog.
     * @param policy The policy to analyze.
     * @return "Okay" if the policy checks out; otherwise an explanation of
     * one or more non-deterministic scenarios is given.
     */
    public static String checkForNonDeterminism(Policy policy)
    {
        // TODO: Fix multi-state matching issues in which multiple separate instance states occur (nodes s obj=x,nodes s obj=y) vs just (nodes s obj=z).
        // TODO: Fix problem of intersecting forall vars where 'from' looks different (e.g., I=I+1)

        // Get forall vars
        // First clear out anything from before
        usedForallVars = new HashSet<String>();
        forallFrom = new HashMap<String, String>();
        forallTo = new HashMap<String, String>();
        // Now recursively search (and possibly modify) forall vars in
        // the policy itself
        extractForallVars(policy.getData().forall, new HashMap<String,String>());

        // Get the list of basic edges (ones not in foralls)
        ArrayList<Edge> edgeList = policy.getData().edge;
        // Append all other edges contained in foralls
        for (int i = 0; i < policy.getData().forall.size(); i++)
        {
            edgeList.addAll(PolicyComponent.extractEdges(policy.getData().forall.get(i)));
        }
        
        // Create the Prolog file
        PrologFile pFile = new PrologFile("temp.pl");
        
        // Include the necessary libraries for linear constraint analysis
        pFile.appendCode(":- use_module(library('clp/bounds')).\n");
        pFile.appendCode(":- use_module([pl-minisat/cnf, pl-minisat/minisat]).\n\n");
        
        // Build a Prolog fact for each edge's nodes tags
        ArrayList<Fact> factList = new ArrayList<Fact>();
        for (int i = 0; i < edgeList.size(); i++)
        {
            Edge edge = edgeList.get(i);
            
            for (int j = 0; j < edge.nodes.size(); j++)
            {
                Nodes nodes = edge.nodes.get(j);
                if (nodes.from.isInteger())
                {
                    // Sanitize all values
                    String name = edge.name.isEmpty() ? "null" : edge.name;
                    String state = nodes.state.isEmpty() ? "null" : nodes.state;
                    String objName = nodes.objName.isEmpty() ? "null" : nodes.objName;
                    factList.add(new Fact("connects",
                            new String[]
                                {
                                    name,
                                    state,
                                    objName,
                                    nodes.from.getString(),
                                    nodes.to.getString()
                                }
                    ));
                }
            }
            
        }
        
        // Write each connection fact for regular (non-forall) edges
        for (int i = 0; i < factList.size(); i++)
            pFile.appendCode(factList.get(i).toString());
        
        // Query type 1: Checks all static (non-forall) edges' from attributes
        if (!edgeList.isEmpty())
        {
            pFile.appendCode("\n" +
                "nondet1(EdgeName1,EdgeName2,To1,To2) :-\n" +
                "   connects(EdgeName1,StateName,_,From,To1),\n" +
                "   connects(EdgeName2,StateName,_,From,To2),\n" +
                "   To1 \\== To2, EdgeName1 @=< EdgeName2,\n" +
                "   \\+disjoint(EdgeName1,EdgeName2).\n" +
                "\n" +
                "disjoint(EdgeName1,EdgeName2) :-\n" +
                "   connects(EdgeName1,S,_,F_1,_),connects(EdgeName2,S,_,F_2,_),F_1\\==F_2.\n\n");
//                "   ((connects(EdgeName1,S2,_,F2_1,_),connects(EdgeName2,S2,_,F2_2,_))->F2_1 =:= F2_2).\n\n");
        }
        
        // Query type 2: Uses linear constraint solving to detect collisions
        // between forall edges and every other edge (forall and otherwise)
        
        // Individual predicates for each edge pair
        HashMap<TripleTuple, String> lcQueryPair = new HashMap<TripleTuple, String>();
        for (int i = 0; i < edgeList.size(); i++)
        {
            ArrayList<Nodes> nodesIList = edgeList.get(i).nodes;
            for (int in = 0; in < nodesIList.size(); in++)
            {
                for (int j = 0; j < edgeList.size(); j++)
                {
//                    if (i != j)
//                    {
                        ArrayList<Nodes> nodesJList = edgeList.get(j).nodes;
                        for (int jn = 0; jn < nodesJList.size(); jn++)
                        {
                            if (nodesIList.get(in).state.equals(nodesJList.get(jn).state))
                            {
                                if (!nodesIList.get(in).from.isInteger())
                                {
                                    TreeSet<String> calledVars = new TreeSet<String>();
                                    Iterator<String> iter = usedForallVars.iterator();
                                    while (iter.hasNext())
                                    {
                                        String var = iter.next();
//                                        if (nodesIList.get(in).from.getString().contains(var))
                                        calledVars.add(var);
//                                        if (nodesJList.get(jn).from.getString().contains(var))
                                        calledVars.add(var);
                                    }
                                    // Set each var's bounds and constraints
                                    iter = calledVars.iterator();
                                    HashMap<String, PairTuple<String,String>> bounds =
                                            new HashMap<String, PairTuple<String,String>>();
                                    while (iter.hasNext())
                                    {
                                        String var = iter.next();
                                        bounds.put(var,
                                                new PairTuple(forallFrom.get(var), forallTo.get(var)));
                                    }
                                    String[] calledVarsArray = new String[calledVars.size()];
                                    calledVars.toArray(calledVarsArray);
                                    // Add in constraints
                                    ArrayList<TripleTuple> constraints = new ArrayList<TripleTuple>();
                                    for (int in2 = 0; in2 < nodesIList.size(); in2++)
                                    {
                                        for (int jn2 = 0; jn2 < nodesJList.size(); jn2++)
                                        {
                                            if (nodesIList.get(in2).state.equals(nodesJList.get(jn2).state))
                                            {
                                                constraints.add(new TripleTuple(
                                                        nodesIList.get(in2).from.getString(),
                                                        "#=",
                                                        nodesJList.get(jn2).from.getString()));
                                            }
                                        }
                                    }
                                    // Now a kluge; to's can contain bad states, and those
                                    // are non-numbers that won't be processed properly by Prolog
                                    String badState = spox.policy.StateExpression.BAD_STATE;
                                    String to1 = nodesIList.get(in).to.getString();
                                    String to2 = nodesJList.get(jn).to.getString();
                                    if (to1.equals(badState) && to2.equals(badState))
                                    {
                                        to1 = "1";
                                        to2 = "1";
                                    }
                                    else if (to1.equals(badState) || to2.equals(badState))
                                    {
                                        to1 = "0";
                                        to2 = "1";
                                    }
                                    constraints.add(new TripleTuple(
                                            to1,
                                            "#\\=",
                                            to2));
                                    TripleTuple[] constraintsArray = new TripleTuple[constraints.size()];
                                    constraints.toArray(constraintsArray);
                                    // Assign name
                                    String functor = "nondet_" + edgeList.get(i).name + "_" + edgeList.get(j).name +
                                                "_" + nodesIList.get(in).state + "_" + nodesIList.get(in).objName;
                                    // Make the rule
                                    LinearConstraintRule lcRule = new LinearConstraintRule(
                                            functor,
                                            calledVarsArray,
                                            bounds,
                                            constraintsArray);
                                    // Write it out to the file
                                    pFile.appendCode(lcRule.toString());
                                    // Add its entry to the mapping for later
                                    if (!lcQueryPair.containsKey(new TripleTuple(edgeList.get(j).name,
                                            edgeList.get(i).name, nodesIList.get(in).state)))
                                    {
                                        lcQueryPair.put(new TripleTuple(edgeList.get(i).name,
                                                edgeList.get(j).name, nodesIList.get(in).state),
                                                lcRule.getCallString());
                                    }
                                }
                            }
                        }
//                    }
                }
            }
        }

        // Write code to file
        try
        { pFile.writeCodeToFile(); }
        catch (IOException ex)
        { ex.printStackTrace(); }
        
        // Initialize the engine and load the prolog file into it for queries
        pFile.initPrologEngine();
        pFile.loadPrologFile();

        // Also load some utilities
        pFile.loadPrologFile("intersect_regex.pl");

        // Now analyze for possible non-deterministic scenarios
        HashSet<PairTuple<String,String>> ndPair =
                new HashSet<PairTuple<String,String>>();

        StringBuilder result = new StringBuilder();
        if (!edgeList.isEmpty() && !factList.isEmpty())
        {
            String basicResult =
                    pFile.queryToString("nondet1(Edge_Name_1,Edge_Name_2,To_1,To_2).", true);
            if (basicResult != null)
            {
                result.append("Possible non-determinism under the following edges:\n\n");
                result.append(basicResult);
                result.append("---------------------------\n\n");
            }

            Hashtable[] basicTable =
                    pFile.runQuery("nondet1(Edge_Name_1,Edge_Name_2,To_1,To_2).", true);
            if (basicTable == null)
                basicTable = new Hashtable[0];
            for (int i = 0; i < basicTable.length; i++)
            {
                ndPair.add(new PairTuple(
                        basicTable[i].get("Edge_Name_1").toString(),
                        basicTable[i].get("Edge_Name_2").toString()));
            }
        }
        
        Iterator<TripleTuple> iter = lcQueryPair.keySet().iterator();
        while (iter.hasNext())
        {
            TripleTuple tuple = iter.next();
            String lcResult = pFile.queryToString(lcQueryPair.get(tuple), true);
            if (lcResult != null)
            {
                result.append("Possible non-determinism between edges " + tuple.first + " and " +
                        tuple.second + " on state " + tuple.third +
                        ", given the following forall conditions:\n\n");
                result.append(lcResult);
                result.append("---------------------------\n\n");
                ndPair.add(new PairTuple(tuple.first,tuple.second));
            }
        }
        
        // Query for satisfiability
        result.append(querySAT(edgeList, ndPair, pFile));

        // Return the final result as a string
        if (result.length() > 0)
            return result.toString();
        else
            return "No non-determinism scenarios detected.\n";
    }

    /**
     * Utility method. Obtains an edge from an edge list by name.
     * @param edgeList
     * @param edgeName
     * @return The edge if it exists; null otherwise.
     */
    private static Edge lookupEdgeByName(
            ArrayList<Edge> edgeList,
            String edgeName)
    {
        for (int i = 0; i < edgeList.size(); i++)
        {
            if (edgeList.get(i).name.equals(edgeName))
                return edgeList.get(i);
        }
        return null;
    }

    /**
     * Given a list of potentially non-deterministic edge pairs, this method
     * will generate SAT queries for the Prolog interface that, if found
     * unsatisfiable, will indicate that non-deterministic scenarios actually
     * exist based upon the underlying pointcuts. The results are returned
     * as a string, to be sent to the user.
     * @param edgeList The list of all edges in the policy.
     * @param ndPair A list of pairs of names of potentially non-deterministic edges.
     * @param pFile Prolog file data.
     * @return Formatted results of the queries.
     */
    private static String querySAT(
            ArrayList<Edge> edgeList,
            HashSet<PairTuple<String,String>> ndPair,
            PrologFile pFile)
    {
        StringBuilder returnString = new StringBuilder();

        // The varMap maps all non-recursive pointcuts to variables for Prolog
        HashMap<Pointcut,String> varMap =
                buildPointcutToVariableMap(edgeList,
                    new HashMap<Pointcut,String>());

        // Add the var mapping for reference
        returnString.append("\nVariable mapping:\n\n");
        returnString.append(varMap.toString() + "\n\n");

        HashMap<Edge,HashSet<Pointcut>> edgePointcuts = new HashMap<Edge,HashSet<Pointcut>>();

        // Obtain constraints for each possible combinatorial set (e.g., triples)
        // based on string matches
        HashMap<HashSet<Pointcut>,SatNode> setConstraintMap =
                Constraint.buildSetConstraintMap(varMap,edgePointcuts,pFile);

        // Check each pair
        Iterator<PairTuple<String,String>> iter = ndPair.iterator();
        while (iter.hasNext())
        {

            StringBuilder edgeSAT = new StringBuilder("cnf(");

            // Get the next edge pair
            PairTuple<String,String> pair = iter.next();
            Edge e1 = lookupEdgeByName(edgeList,(String)pair.first);
            Edge e2 = lookupEdgeByName(edgeList,(String)pair.second);
            Pointcut p1 = e1.label.pointcut;
            Pointcut p2 = e2.label.pointcut;

            HashSet<Pointcut> p1VarSet = new HashSet<Pointcut>();
            HashSet<Pointcut> p2VarSet = new HashSet<Pointcut>();

            // Build the individual SAT cases
            SatNode SAT1 = buildSATTree(p1,varMap,p1VarSet);
            SatNode SAT2 = buildSATTree(p2,varMap,p2VarSet);
            returnString.append(pair.first + ": " + SAT1.toString() + "\n");
            returnString.append(pair.second + ": " + SAT2.toString() + "\n");

            edgePointcuts.put(e1, p1VarSet);
            edgePointcuts.put(e2, p2VarSet);

            // Now generate constraints based upon each atomic pointcut pair
            HashMap<Pointcut,String> pairVarMap = new HashMap<Pointcut,String>();
            Iterator<Pointcut> pIter = p1VarSet.iterator();
            while (pIter.hasNext())
            {
                Pointcut p = pIter.next();
                if (!pairVarMap.containsKey(p))
                    pairVarMap.put(p, varMap.get(p));
            }
            pIter = p2VarSet.iterator();
            while (pIter.hasNext())
            {
                Pointcut p = pIter.next();
                if (!pairVarMap.containsKey(p))
                    pairVarMap.put(p, varMap.get(p));
            }
            HashMap<PairTuple<Pointcut,Pointcut>,SatNode> pairConstraintMap =
                    Constraint.buildPairConstraintMap(pairVarMap, pFile);

            // Generate the query and run it
            edgeSAT.append(
                    SAT1.toString() + "*" + SAT2.toString());
            Iterator<SatNode> constraintIter = pairConstraintMap.values().iterator();
            while (constraintIter.hasNext())
            {
                SatNode constraint = constraintIter.next();
                if (!constraint.name.equals("1"))
                    edgeSAT.append("*" + constraint);
            }

            Iterator<HashSet<Pointcut>> setConstraintIter =
                    setConstraintMap.keySet().iterator();
            while (setConstraintIter.hasNext())
            {
                SatNode constraint = setConstraintMap.get(setConstraintIter.next());
                edgeSAT.append("*" + constraint);
            }

            edgeSAT.append(",Cnf),solve(Cnf).");
            String result = pFile.queryToString(edgeSAT.toString(), true);

            // Generate output

            if (result == null)
            {
                returnString.append(edgeSAT.toString() + "\n" +
                        "Edges " + pair.first + " and " + pair.second +
                        " are not non-deterministic.\n\n");
            }
            else
            {
                returnString.append("Edges " + pair.first + " and " + pair.second +
                        " are non-deterministic.\n\n");
                returnString.append(edgeSAT.toString() + "\n" + result + "\n");
            }
        }

        return returnString.toString();
    }

    /**
     * Recursively builds a SAT tree from a given pointcut and returns the
     * topmost node. All recursive pointcuts are represented as operations:<br />
     * <br />
     * AND -> '*'<br />
     * OR  -> '+'<br />
     * NOT -> '-'<br />
     * <br />
     * Note that empty AND and OR blocks are assigned a leaf node of '1', or true.
     * @param p The pointcut to be considered.
     * @param varMap A mapping from individual pointcuts to their variables.
     *               Meant to handle duplicate or intersecting pointcuts,
     *               which should all share the same variable name.
     * @param varSet The set of all variables which appear in the tree.
     * @return The topmost node of the tree.
     */
    private static SatNode buildSATTree(
            Pointcut p,
            HashMap<Pointcut,String> varMap,
            HashSet<Pointcut> varSet)
    {
        if (p instanceof And)
        {
            And and = (And)p;

            ArrayList<Pointcut> pList = and.p;

            // Construct a tree to represent a (possibly huge) conjunction
            SatNode returnNode;
            if (pList.size() == 1)
                returnNode = buildSATTree(pList.get(0),varMap,varSet);
            else if (pList.size() > 1)
            {
                And a = new And();
                ArrayList<Pointcut> aList = new ArrayList<Pointcut>();
                aList.addAll(pList.subList(1, pList.size()));
                a.p = aList;
                returnNode = new SatBinaryParent("*",
                        buildSATTree(pList.get(0),varMap,varSet),
                        buildSATTree(a, varMap, varSet));
            }
            else
            {
                returnNode = new SatLeaf("1");
            }
            return returnNode;
        }
        else if (p instanceof Or)
        {
            Or and = (Or)p;

            ArrayList<Pointcut> pList = and.p;

            // Construct a tree to represent a (possibly huge) disjunction
            SatNode returnNode = new SatLeaf("1");
            if (pList.size() == 1)
                returnNode = buildSATTree(pList.get(0),varMap,varSet);
            else if (pList.size() > 1)
            {
                Or o = new Or();
                ArrayList<Pointcut> oList = new ArrayList<Pointcut>();
                oList.addAll(pList.subList(1, pList.size()));
                o.p = oList;
                returnNode = new SatBinaryParent("+",
                        buildSATTree(pList.get(0),varMap,varSet),
                        buildSATTree(o, varMap, varSet));
            }
            return returnNode;
        }
        else if (p instanceof Not)
        {
            Not not = (Not)p;
            SatNode returnNode = new SatUnaryParent("-",
                    buildSATTree(not.p, varMap,varSet));
            return returnNode;
        }
        else if (p != null)
        {
            varSet.add(p);
            return new SatLeaf(varMap.get(p));
        }
        else
            return new SatLeaf("1");
    }

    private static HashMap<Pointcut, String> buildPointcutToVariableMap(
            ArrayList<Edge> edgeList,
            HashMap<Pointcut,String> varMap)
    {
        for (int i = 0; i < edgeList.size(); i++)
        {
            Edge edge = edgeList.get(i);
            Pointcut p = edge.label.pointcut;
            ArrayList<Pointcut> singlePList = new ArrayList<Pointcut>();
            singlePList.add(p);
            buildPointcutToVariableMap(varMap,singlePList);
        }
        return varMap;
    }

    private static HashMap<Pointcut, String> buildPointcutToVariableMap(
            HashMap<Pointcut,String> varMap,
            ArrayList<Pointcut> pList)
    {
        for (int i = 0; i < pList.size(); i++)
        {
            Pointcut p = pList.get(i);
            if (p instanceof And)
            {
                buildPointcutToVariableMap(varMap,((And)p).p);
            }
            else if (p instanceof Or)
            {
                buildPointcutToVariableMap(varMap,((Or)p).p);
            }
            else if (p instanceof Not)
            {
                ArrayList<Pointcut> singlePList = new ArrayList<Pointcut>();
                singlePList.add(((Not)p).p);
                buildPointcutToVariableMap(varMap,singlePList);
            }
            else
            {
                if (!varMap.containsKey(p))
                {
                    String varName = generateUniqueSATVar(varMap);
                    varMap.put(p, varName);
                }
            }
        }
        return varMap;
    }

    public static enum Alphabet {A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z};

    /**
     * Automatically generates a unique variable to use in a SAT boolean sentence.
     * @param varMap The mapping of pointcut commands to their variable names.
     * @return A unique variable name.
     */
    public static String generateUniqueSATVar(HashMap<Pointcut,String> varMap)
    {
        int alpha = 0;
        int append = 0;
        boolean useAppend = false;
        while (true)
        {
            String s = Alphabet.values()[alpha].toString() +
                    (useAppend ? ++append : "");
            if (varMap.values().contains(s))
            {
                if (alpha++ >= Alphabet.values().length)
                    alpha = 0;
            }
            else
            {
                return s;
            }
        }
    }
    
    public static SatNode eliminateExtraVars(SatNode tree, HashSet<String> varSet)
    {
        if (tree instanceof SatBinaryParent)
        {
            SatBinaryParent bTree = (SatBinaryParent)tree;
            SatNode left = eliminateExtraVars(bTree.left,varSet);
            SatNode right = eliminateExtraVars(bTree.right,varSet);
            if (left==null)
                return right;
            else if (right==null)
                return left;
            else
            {
                bTree.left = left;
                bTree.right = right;
                return bTree;
            }
        }
        else if (tree instanceof SatUnaryParent)
        {
            SatUnaryParent uTree = (SatUnaryParent)tree;
            SatNode child = eliminateExtraVars(uTree.child,varSet);
            if (child==null)
                return null;
            else
            {
                uTree.child = child;
                return uTree;
            }
        }
        else
        {
            if (varSet.contains(tree.name))
                return tree;
            else
                return null;
        }
    }
    
    /**
     * Extracts all forall var names from a list of foralls. If duplicates are
     * encountered, they are changed within the policy object to something
     * unique. This greatly simplifies linear constraint solving issues.
     * Note that it also converts all var names to uppercase for Prolog's sake.
     * This method also populates the from and to mappings.
     * <br /><br />
     * NOTE: This method has the side effect of potentially modifying the
     * policy. Use with care.
     * 
     * @param forallList The Forall list to analyze (and possibly modify).
     * @param oldToNewNameMap A mapping from old to new forall var names.
     * @returns The final oldToNewMap.
     */
    private static HashMap<String,String> extractForallVars(
            ArrayList<Forall> forallList,
            HashMap<String,String> oldToNewNameMap)
    {
        for (int i = 0; i < forallList.size(); i++)
        {
            Forall forall = forallList.get(i);
            String originalName = forall.var;
            forall.var = forall.var.toUpperCase();
            // If it's been taken, rename it
            String tempName = forall.var;
            int count = 2;
            while (usedForallVars.contains(tempName))
            {
                tempName = forall.var + count;
                count++;
            }
            forall.var = tempName;
            usedForallVars.add(forall.var);
            forallFrom.put(forall.var, Integer.toString(forall.from));
            forallTo.put(forall.var, Integer.toString(forall.to));
            
            // Convert any underlying edges and foralls as needed
            oldToNewNameMap.put(originalName, forall.var);
            for (int j = 0; j < forall.edge.size(); j++)
            {
                Edge edge = forall.edge.get(j);
                for (int k = 0; k < edge.nodes.size(); k++)
                {
                    Nodes nodes = edge.nodes.get(k);
                    Iterator<String> iter = oldToNewNameMap.keySet().iterator();
                    while (iter.hasNext())
                    {
                        String oldName = iter.next();
                        if (nodes.from.getString().contains(oldName))
                        {
                            nodes.from = new StateExpression(
                                    nodes.from.getString().replace(
                                        oldName, oldToNewNameMap.get(oldName)));
                        }
                        if (nodes.to.getString().contains(oldName))
                        {
                            nodes.to = new StateExpression(
                                    nodes.to.getString().replace(
                                        oldName, oldToNewNameMap.get(oldName)));
                        }
                    }
                }
            }
            
            // NOTE: This will need to be changed once froms and tos get
            // converted to strings (like StateExpressions).
            // Right now this is only a stub.
            for (int j = 0; j < forall.forall.size(); j++)
            {
                Forall fa = forall.forall.get(j);
                Iterator<String> iter = oldToNewNameMap.keySet().iterator();
                while (iter.hasNext())
                {
                    String oldName = iter.next();
                    if (Integer.toString(fa.from).contains(oldName))
                    {
                    }
                    if (Integer.toString(fa.to).contains(oldName))
                    {
                    }
                }
            }

            // Now recursively extract and modify any underlying foralls
            extractForallVars(forall.forall, oldToNewNameMap);
        }
        
        return oldToNewNameMap;
    }
}
